type scanline = {
    sid : int;
    bx : int;
    ex : int;
    y : int;
    mutable neighbour : scanline list;
  }
type margin =
    Safe
  | Left
  | Right
  | Left_margin
  | Right_margin

let is_in s x y = s.bx <= x && x <= s.ex && y = s.y

let margin s x =
  if x < s.bx then Left else
  if s.ex < x then Right else
  if x < s.bx + min 2 ((s.ex - s.bx) / 2) then Left_margin else
  if s.ex - min 2 ((s.ex - s.bx) / 2) < x then Right_margin
  else Safe

class map (width, height) mapstring =
  let get x y = String.get mapstring ((width + 2) * y + x) in
  let passable x y = let c = get x y in c = '.' || c = '@' in

  let home =
    let home = ref [] in
    for i = 0 to String.length mapstring - 1 do
      if String.get mapstring i = '@' then
	home := (i mod (width + 2), i / (width + 2))::!home;
    done;
    !home in

  let smap = Array.make_matrix (width + 1) (height + 1) None in

  let sarray = 
    let cid = ref 0 in
    let slist = ref [] in

    let rec loop x y =
      match smap.(x).(y) with
	Some s -> s
      | None ->
	  let rec bx x = if passable (x - 1) y then bx (x - 1) else x in
	  let rec ex x = if passable (x + 1) y then ex (x + 1) else x in
	  let this = { sid = !cid; bx = bx x; ex = ex x; y = y; neighbour = [] } in
	  cid := !cid + 1;
	  slist := this::!slist;

	  for x = this.bx to this.ex do
	    smap.(x).(y) <- Some this
	  done;

	  let rec skip_inpassable x dy =
	    if x > this.ex then [] else
	    if passable x (y + dy) then
	      loop x (y + dy)::skip_passable (x + 1) dy
	    else
	      skip_inpassable (x + 1) dy
	  and skip_passable x dy =
	    if x > this.ex then [] else
	    if passable x (y + dy) then skip_passable (x + 1) dy
	    else skip_inpassable (x + 1) dy in

	  this.neighbour <- skip_inpassable this.bx (-1) @ skip_inpassable this.bx 1;
	  this in
    for x = 1 to width do
	for y = 1 to height do
	  if passable x y && smap.(x).(y) = None then ignore (loop x y)
	done
    done;
    Array.of_list (List.rev !slist) in

  let sarray_size = Array.length sarray in

  let neighbour_id id =
    let s = sarray.(id) in
    List.map (fun x -> x.sid) s.neighbour in

  let distance id1 x id2 =
    if x < sarray.(id2).bx then sarray.(id2).bx - x + 1 else
    if sarray.(id2).ex < x then x - sarray.(id2).ex + 1 else
    1 in

  object (self)
    method debug () =
      Global.debug (Printf.sprintf "Smap.map width=%d height=%d" width height);
      for i = 0 to height + 1 do
	Global.debug (Printf.sprintf "%s" (String.sub mapstring (i * (width + 2)) (width + 2)))
      done;
      List.iter (fun (x, y) -> Global.debug (Printf.sprintf "home at %d, %d" x y)) home;
      Array.iter (fun s -> Global.debug (Printf.sprintf "scanline (%d) bx:%d ex:%d y:%d"
					  s.sid s.bx s.ex s.y)) sarray
    method width = width
    method height = height
    method mapstring = mapstring
    method home = home

    method get_scanline id = sarray.(id)

    method dijkstra_raw bx by ex ey =
      let bid, eid = match smap.(bx).(by), smap.(ex).(ey) with
	Some bs, Some es -> bs.sid, es.sid
      | None, Some es -> failwith "Source Unreachable"
      |	Some bs, None -> failwith "Destination Unreachable"
      |	None, None -> failwith "S & D Unreachable" in
      if bid = eid then [sarray.(bid)], abs(ex - bx) else
      let dist = Array.create sarray_size 1000000 in (* TODO *)
      let prev = Array.create sarray_size (-1) in

      let min_edge = Global.min_of (fun (id1, x, id2) -> dist.(id2)) in

      let rec loop s e =
	let (m, cx, n), d = min_edge e in
	prev.(n) <- m;
	let s' = n::s in
	let cx' =
	  if cx < sarray.(n).bx then sarray.(n).bx else
	  if sarray.(n).ex < cx then sarray.(n).ex else
	  cx in
	if n = eid then
	  dist.(eid) <- d + abs(cx' - ex)
	else
	  let e' = List.fold_right begin fun x l ->
	    if List.mem x s' then l
	    else begin
	      if dist.(x) > d then dist.(x) <- (d + distance n cx' x);
	      (n, cx', x)::l
	    end
	  end (neighbour_id n) (List.filter (fun (_, _, x) -> x <> n) e) in
	  loop s' e' in

      dist.(bid) <- 0;
      let e = List.map begin fun x ->
	dist.(x) <- (distance bid bx x); (bid, bx, x)
      end (neighbour_id bid) in
      loop [bid] e;
      let rec trace id r =
	if id = bid then sarray.(id)::r else trace prev.(id) (sarray.(id)::r) in
      trace eid [], dist.(eid)

    val dijkstra_memo = Hashtbl.create 101

    method dijkstra bx by ex ey =
      try Hashtbl.find dijkstra_memo (bx, by, ex, ey)
      with Not_found ->
	let result = self#dijkstra_raw bx by ex ey in
	Hashtbl.replace dijkstra_memo (bx, by, ex, ey) result;
	result
  end
